/*
 * uidls.c
 *
 * This file is part of mpop, a POP3 client.
 *
 * Copyright (C) 2005
 * Martin Lambers <marlam@marlam.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   mpop is released under the GPL with the additional exemption that
 *   compiling, linking, and/or using OpenSSL is allowed.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
extern int errno;

#include "uidls.h"
#include "list.h"
#include "xalloc.h"

/* a line from an UIDLs file must fit into a buffer of this size: */
#define UIDLS_LINEBUFSIZE 501


/*
 * uidl_new()
 *
 * see uidls.h
 */

uidl_t *_uidl_new(void)
{
    uidl_t *uidl = xmalloc(sizeof(uidl_t));
    uidl->hostname = NULL;
    uidl->user = NULL;
    uidl->n = 0;
    uidl->uidv = NULL;
    return uidl;
}

uidl_t *uidl_new(const char *hostname, const char *user)
{
    uidl_t *uidl = _uidl_new();
    uidl->hostname = xstrdup(hostname);
    uidl->user = xstrdup(user);
    return uidl;
}


/*
 * uidl_free()
 *
 * see uidls.h
 */

void uidl_free(void *u)
{
    uidl_t *uidl = u;
    long i;
    
    if (uidl)
    {
	free(uidl->hostname);
	free(uidl->user);
        if (uidl->uidv)
	{
	    for (i = 0; i < uidl->n; i++)
	    {
		free(uidl->uidv[i]);
	    }
	    free(uidl->uidv);
	}
	free(uidl);
    }
}


/*
 * uidl_find()
 *
 * see uidls.h
 */

uidl_t *find_uidl(list_t *uidl_list, const char *hostname, const char *user)
{
    uidl_t *tmp;
    uidl_t *uidl = NULL;
    list_t *lp = uidl_list;

    while (!list_is_empty(lp))
    {
	lp = lp->next;
	tmp = lp->data;
	if (strcmp(tmp->hostname, hostname) == 0 
		&& strcmp(tmp->user, user) == 0)
	{
	    uidl = tmp;
	    break;
	}
    }

    return uidl;
}


/*
 * uidls_read()
 *
 * see uidls.h
 */

int _uidls_qsort_strcmp(const void *a, const void *b)
{
    return strcmp(*(char **)a, *(char **)b);
}

int uidls_read(const char *filename, list_t **uidl_list, char *errstr)
{
    FILE *f;
    char line[UIDLS_LINEBUFSIZE];
    uidl_t *uidl;
    list_t *lp;
    char *p, *q;
    long i;
    int sorted;
    
    *uidl_list = list_new();
    if (!(f = fopen(filename, "r")))
    {
	return UIDLS_EOK;
    }
    lp = *uidl_list;
    uidl = NULL;
    while (fgets(line, (int)sizeof(line), f))
    {
	if ((p = strchr(line, '\n')))
	{
	    *p = '\0';
	}
	else if (strlen(line) == UIDLS_LINEBUFSIZE - 1)
	{
	    (void)fclose(f);
	    snprintf(errstr, errstr_bufsize,
		    "%s: line longer than %d characters", 
		    filename, UIDLS_LINEBUFSIZE - 1);
	    list_xfree(*uidl_list, uidl_free);
	    return UIDLS_EFORMAT;
	}
	if (line[0] == '#')
	{
	    /* a comment */
	    continue;
	}
	else if (line[0] != ' ')
	{
	    (void)fclose(f);
	    snprintf(errstr, errstr_bufsize,
		    "%s: UID without a list", filename);
	    list_xfree(*uidl_list, uidl_free);
	    return UIDLS_EFORMAT;
	}
	else
	{
	    /* expecting " number hostname user" */
	    uidl = _uidl_new();
	    uidl->n = strtol(line + 1, &p, 10);
	    if (uidl->n < 1 || uidl->n == LONG_MAX 
		    || (p == line + 1) || *p != ' ')
	    {
		snprintf(errstr, errstr_bufsize, 
			"%s: invalid number of UIDs", filename);
		(void)fclose(f);
		uidl->n = 0;
		list_xfree(*uidl_list, uidl_free);
		return UIDLS_EFORMAT;
	    }
	    uidl->uidv = xmalloc(uidl->n * sizeof(char *));
	    for (i = 0; i < uidl->n; i++)
	    {
		uidl->uidv[i] = NULL;
	    }		    
	    p++;
	    if (!(q = strchr(p, ' ')) || (q == p))
	    {
		snprintf(errstr, errstr_bufsize, 
			"%s: invalid or missing host name", filename);
		(void)fclose(f);
		list_xfree(*uidl_list, uidl_free);
		return UIDLS_EFORMAT;
	    }
	    uidl->hostname = xmalloc((q - p + 1) * sizeof(char));
	    strncpy(uidl->hostname, p, (size_t)(q - p));
	    uidl->hostname[q - p] = '\0';
	    p = q + 1;
	    if (*p == '\0' || *p == ' ' || strchr(p, ' '))
	    {
		snprintf(errstr, errstr_bufsize, 
			"%s: invalid or missing user name", filename);
		(void)fclose(f);
		list_xfree(*uidl_list, uidl_free);
		return UIDLS_EFORMAT;
	    }
	    uidl->user = xstrdup(p);
	    list_insert(lp, uidl);
	    lp = lp->next;
	    sorted = 1;
	    /* now read the list of UIDs */
	    for (i = 0; i < uidl->n; i++)
	    {
		if (!fgets(line, (int)sizeof(line), f))
		{
		    snprintf(errstr, errstr_bufsize, 
			    "%s: too few UIDs for user %s, host %s", 
			    filename, uidl->user, uidl->hostname);
		    (void)fclose(f);
		    list_xfree(*uidl_list, uidl_free);
		    return UIDLS_EFORMAT;
		}
		if ((p = strchr(line, '\n')))
		{
		    *p = '\0';
		}
		else if (strlen(line) == UIDLS_LINEBUFSIZE - 1)
		{
		    snprintf(errstr, errstr_bufsize,
			    "%s: line longer than %d characters", 
			    filename, UIDLS_LINEBUFSIZE - 1);
		    (void)fclose(f);
		    list_xfree(*uidl_list, uidl_free);
		    return UIDLS_EFORMAT;
		}
		if (line[0] == '#')
		{
		    /* a comment */
		    continue;
		}
		else if (line[0] == ' ')
		{
		    snprintf(errstr, errstr_bufsize,
			    "%s: too few UIDs for user %s, host %s", 
			    filename, uidl->user, uidl->hostname);
		    (void)fclose(f);
		    list_xfree(*uidl_list, uidl_free);
		    return UIDLS_EFORMAT;
		}
		uidl->uidv[i] = xstrdup(line);
		if (i >= 1 && sorted)
		{
		    if (strcmp(uidl->uidv[i - 1], uidl->uidv[i]) > 0)
		    {
			sorted = 0;
		    }
		}
	    }
	    if (!sorted)
    	    {
		/* this should only happen when we read an UIDLS written by an
		 * older version. */
		qsort(uidl->uidv, (size_t)uidl->n, sizeof(char *), 
			_uidls_qsort_strcmp);
	    }
	}
    }
    if (ferror(f))
    {
	(void)fclose(f);
	list_xfree(*uidl_list, uidl_free);
	snprintf(errstr, errstr_bufsize, "%s: input error", filename);
	return UIDLS_EIO;
    }
    (void)fclose(f);

    return UIDLS_EOK;
}


/*
 * uidls_write()
 *
 * see uidls.h
 */

int uidls_write(const char *filename, list_t *uidl_list, char *errstr)
{
    FILE *f;
    list_t *lp;
    uidl_t *uidl;
    long i;

    if (!(f = fopen(filename, "w")))
    {
	goto error_exit;
    }
    
    if (fprintf(f, 
		"# This file was generated by %s version %s.\n"
		"#\n"
		"# Lines starting with a space begin a new UID list and have\n"
		"# the following format: \" number-of-uids hostname user\".\n"
		"# The list of UIDs follows in ascending order: "
			"one UID per line, no spaces.\n"
		"#\n",
		PACKAGE_NAME, PACKAGE_VERSION) < 0)
    {
	goto error_exit;
    }
    
    lp = uidl_list;
    while (!list_is_empty(lp))
    {
	lp = lp->next;
	uidl = lp->data;
	if (uidl->n > 0)
	{
	    if (fprintf(f, " %ld %s %s\n", uidl->n, uidl->hostname, uidl->user)
		    < 0)
	    {
		goto error_exit;
	    }
	    for (i = 0; i < uidl->n; i++)
	    {
		if (fputs(uidl->uidv[i], f) == EOF 
		     	|| fputc((unsigned char)'\n', f) == EOF)
		{
		    goto error_exit;
		}
	    }
	}
    }
    
    (void)fclose(f);
    return UIDLS_EOK;

error_exit:
    if (f)
    {
	(void)fclose(f);
    }
    snprintf(errstr, errstr_bufsize, "cannot write %s: %s", 
	    filename, strerror(errno));
    return UIDLS_EIO;
}
